---@diagnostic disable: deprecated
local key = ModPath .. '	' .. RequiredScript
if _G[key] then return else _G[key] = true end

_G.D_RUNNIN = _G.D_RUNNIN or {}
D_RUNNIN._path = ModPath
D_RUNNIN._data_path = SavePath .. 'D_RUNNIN.json'

blt.xaudio.setup()
XAudio._base_gains.bazinga = 1
XAudio.Source.BAZINGA = "bazinga"

D_RUNNIN.settings = {
    time_to_sweet = 600, -- 600
    chances = {
        mutate_interlude = 0.05,
        mutate_interlude_start = 0.15,
        mutate_switch_toggle = 0.33,
        mutate_switch_toggle_target = 0.25,
        mutate_switch_toggle_non_target = 0.75,
    },
    ping = 666,
    target_ids = {
        ["D_RUNNIN2"] = true,
        ["D_RUNNIN3"] = true,
    },
    target_sources = {
        "dc7dc6383128",
        "defa8da952ef",
        "b829bfe5155d",
    },
}
D_RUNNIN.last_ping = 0
D_RUNNIN.candy_jar = {}
D_RUNNIN.noncandy_jar = {}
D_RUNNIN.did_first_run = false


--#region debug local
--[[
D_RUNNIN.settings.known_events = {
    "music_heist_setup",
    "music_heist_control",
    "music_heist_anticipation",
    "music_heist_assault",
    "suspense_1",
    "suspense_2",
    "suspense_3",
    "suspense_4",
    "suspense_5",
    "menu_music",
    "stop_all_music",
}
D_RUNNIN.settings.print_depth = 5
D_RUNNIN.think_dump = true
D_RUNNIN.debug = false
D_RUNNIN.debug_chat = false
local function table_contains(table, element)
    for _, value in pairs(table) do
        if value == element then
        return true
        end
    end
    return false
end

local function format_round_digits(num, idp)
    local mult = 10 ^ ( idp or 0 )
    num = math.floor( num * mult + 0.5 ) / mult
    return (idp or 0) and tostring(num) or string.format("%."..(idp or 0).."f", num):gsub("%.?0+$", "")
end

local chat_log = function(...)
    if not D_RUNNIN.debug_chat then return end
    local argv = {...}
    for i = 1, #argv, 1 do
        argv[i] = tostring(argv[i])
    end
    local msg = table.concat(argv, " ")

    log("D_RUNNIN_CL: " .. msg)
    if managers.chat then
        managers.chat:_receive_message(1,"D_RUNNIN", msg, Color("FF00FF"))
    end
end

local tracer = function(name, ...)
    if not D_RUNNIN.debug then return end
    local ext = {...}
    local msg = name or ('[sorry nothing]' .. table.concat(ext, ";"))

    if not table_contains(D_RUNNIN.settings.known_events, msg) then
        chat_log('music_event ' .. msg)
    end
end
]]
--#endregion debug local

--#region local functions
local function easeInExpo(start, nd, val)
    nd = nd - start
    return nd * math.pow(2, 10 * (val / 1 - 1)) + start;
end

local function easeOutCirc(start, nd, val)
	val = val - 1;
	nd = nd - start;
	return nd * math.pow(1 - val * val, 1/2) + start
end


local rhythm = function(x, phase, space)
    local space = 1+(space or 0)
    return -1 * math.sin( math.rad( ((2*math.pi*x) / phase) * (1/space) ) )
end


local function bare_event(event)
    return string.gsub(event, "music_heist_", "")
end

local function get_current_track(track)
    local current_track = track or (Global.music_manager and Global.music_manager.current_track)
    return (D_RUNNIN.settings.target_ids[ current_track ] and current_track) or false
end

local function get_time()
    return (managers.game_play_central and managers.game_play_central:get_heist_timer()) or 0
end
--#endregion local functions

function D_RUNNIN:roll_chance(chance_name)
    local beat = -math.huge
    local peak = -math.huge
    local time = get_time()
    if self.settings.chances[ chance_name ] ~= nil then
        peak = self.settings.chances[ chance_name ]
        if time < self.settings.time_to_sweet then
            beat = easeInExpo(
                0,
                peak,
                math.min((time/self.settings.time_to_sweet), 1)
            )
        else
            -- go ballistic
            local min = (time-self.settings.time_to_sweet)
            local max = (self.settings.time_to_sweet * 10)
            local ppeak = math.pow(peak, 1/2)
            beat = easeInExpo(
                peak,
                ppeak,
                math.min(min/max, 1)
            )
            -- chat_log('&'..chance_name,  format_round_digits(min,2) .. '/' .. format_round_digits(max,2) .. '|' .. format_round_digits(peak,2) .. '/' .. format_round_digits(ppeak,2) .. '=' .. format_round_digits(beat,2))
        end
    end
    local roll = math.random()
    local success = (roll < beat)
    -- chat_log(success and '[OK]' or '[NG]', '@'..chance_name,  format_round_digits(roll,2) .. '/' .. format_round_digits(beat,2) .. '|' .. format_round_digits(peak,2))
    return success
end

function D_RUNNIN:ping()
    local bazinga = XAudio.Source:new(XAudio.Buffer:new(self._path .. "__tmp_oggs/749658afb319"))
    bazinga._type = XAudio.Source.BAZINGA
    bazinga._gain = 1
    bazinga._raw_gain = 1
end

function D_RUNNIN:IsTargetSource(input_path)
    for _, target_source in pairs(self.settings.target_sources) do
        local target_found = not not (string.find(input_path, target_source))
        if target_found then
            return true
        end
    end
    return false
end

function D_RUNNIN:GetInterlude(is_start)
    local desired_interlude = (is_start and "interlude_start") or "interlude"

    local current_track = get_current_track()
    if not current_track then return end
    local mus = BeardLib.MusicMods and BeardLib.MusicMods[ current_track ] or nil

    if mus and mus.events then
        return mus.events[ desired_interlude ]
    end
end

function D_RUNNIN:setup_track(track)
    -- initial setup
    if not track.was_setup then
        track.weight = tonumber(track.weight or 1)
        track.orig_weight = track.weight
        track.weight_offset = math.random()

        track.was_setup = true
    end

    return track
end

function D_RUNNIN:setup_tracks()
    local current_track = get_current_track()
    local mus = BeardLib.MusicMods and BeardLib.MusicMods[current_track] or nil
    if not mus then return end

    for _, event in pairs(mus.events) do
        ---@todo: BeardLib sets this even though i never said to
        event.allow_switch = nil
        for _, track in pairs(event.tracks) do
            self:setup_track(track)
        end
    end
end

-- Recursive deep_clone http://lua-users.org/wiki/CopyTable
local function deepcopy(orig, copies)
    copies = copies or {}
    local orig_type = type(orig)
    local copy
    if orig_type == 'table' then
        if copies[orig] then
            copy = copies[orig]
        else
            copy = {}
            copies[orig] = copy
            for orig_key, orig_value in next, orig, nil do
                if orig_key == 'module' then
                    copy[deepcopy(orig_key, copies)] = orig_value
                else
                    copy[deepcopy(orig_key, copies)] = deepcopy(orig_value, copies)
                end
            end
            setmetatable(copy, deepcopy(getmetatable(orig), copies))
        end
    else -- number, string, boolean, etc
        copy = orig
    end
    return copy
end

function D_RUNNIN:JarCandy()
    if self.did_first_run then return end
    local current_track = get_current_track()
    if not current_track then return end

    local mus = BeardLib.MusicMods and BeardLib.MusicMods[current_track] or nil

    if mus then
        self.candy_jar = {}
        self.noncandy_jar = {}
        self.candy_jar[ current_track ] = mus
        -- log('D_RUNNIN: successfully jarred the candy')
    -- else
        -- log('D_RUNNIN: failed to jar the candy')
    end
end

-- most jars actually are infinite
-- that's a little tidbit they don't tell you
function D_RUNNIN:DumpCandyJar(current_track)
    local mus = self.candy_jar[ current_track ]

    -- log('dumping candy jar ...')
    -- Utils.PrintTable(mus, self.settings.print_depth)

    if mus then
        BeardLib.MusicMods[ current_track ] = deepcopy(mus, self.noncandy_jar)
        -- log('D_RUNNIN: successfully dumped the candy jar')
    -- else
        -- log('D_RUNNIN: failed to dump the candy jar')
    end
end

function D_RUNNIN:mutate_interlude_start(event)
    -- log('finding interlude_start')
    if D_RUNNIN:roll_chance('mutate_interlude_start') then
        local current_track = get_current_track()
        if not current_track then return end

        local mus = (BeardLib.MusicMods and BeardLib.MusicMods[current_track]) or nil
        local mus_event = (mus and mus.events and mus.events[event]) or nil
        -- Utils.PrintTable(mus_event, self.settings.print_depth)

        if mus_event then
            local interlude_event = self:GetInterlude(true)
            -- Utils.PrintTable(interlude_event, self.settings.print_depth)
            if interlude_event ~= nil then
                local interlude_track = interlude_event.tracks[ MusicManager:pick_track_index(interlude_event.tracks) ]
                for _, mus_track in pairs(mus_event.tracks) do
                    mus_track.start_source = interlude_track.source
                end
            -- else
                -- log('D_RUNNIN: interlude_startn\'t')
            end
        end
    end
end

function D_RUNNIN:mutate_interlude(event)
    -- log('finding interlude')
    if D_RUNNIN:roll_chance('mutate_interlude') then
        local current_track = get_current_track()
        if not current_track then return end

        local mus = (BeardLib.MusicMods and BeardLib.MusicMods[current_track]) or nil
        local mus_event = (mus and mus.events and mus.events[event]) or nil
        -- Utils.PrintTable(mus_event, self.settings.print_depth)

        if mus_event then
            local interlude_event = self:GetInterlude()
            -- Utils.PrintTable(interlude_event, self.settings.print_depth)
            if interlude_event ~= nil then
                local interlude_track = interlude_event.tracks[ MusicManager:pick_track_index(interlude_event.tracks) ]
                for _, mus_track in pairs(mus_event.tracks) do
                    mus_track.source = interlude_track.source
                end
            -- else
                -- log('D_RUNNIN: interluden\'t')
            end
        end
    end
end

function D_RUNNIN:mutate_switch_toggle(event)
    --[[
        we used to do this in think, and maybe it has an effect prior to
        being called via post_event, but _switch_at_end i think is
        computed ahead of time so we can lighten the think function

        but it would be really cool to toggle allow_switch after a
        track part has looped at least once during the same segment
    ]]
    -- log('switch toggle')
    if D_RUNNIN:roll_chance('mutate_switch_toggle') then
        local current_track = get_current_track()
        if not current_track then return end

        local mus = (BeardLib.MusicMods and BeardLib.MusicMods[current_track]) or nil
        local mus_event = (mus and mus.events and mus.events[event]) or nil
        -- Utils.PrintTable(mus_event, self.settings.print_depth)

        if mus_event then
            local mus_track = mus_event.tracks[ MusicManager:pick_track_index(mus_event.tracks) ]

            local target_found = self:IsTargetSource(mus_track.source.path)
            if (target_found and D_RUNNIN:roll_chance('mutate_switch_toggle_target')) or
                (not target_found and D_RUNNIN:roll_chance('mutate_switch_toggle_non_target')) then
                if mus_track.allow_switch ~= nil then
                    mus_track.allow_switch = nil
                else
                    mus_track.allow_switch = true
                end
            end
        end
    end
end

function D_RUNNIN:mutate_ambient_weight(event)
    local current_track = get_current_track()
    if not current_track then return end

    local mus = (BeardLib.MusicMods and BeardLib.MusicMods[current_track]) or nil
    local mus_event = (mus and mus.events and mus.events[event]) or nil

    if mus_event then
        for _, mus_track in pairs(mus_event.tracks) do
            local target_found = self:IsTargetSource(mus_track.source.path)
            if not target_found then
                local goal_weight = mus_track.orig_weight
                goal_weight = 1 + rhythm(get_time(), mus_track.orig_weight)
                mus_track.weight = easeOutCirc(0, goal_weight, math.min(get_time()/self.settings.time_to_sweet, 1))
                -- log('D_RUNNIN: touching ' .. mus_track.source.path  .. ' at ' .. tostring(mus_track.weight) .. ' for ' .. tostring(get_time()))
            end
        end
    end
end

function D_RUNNIN:mutate_target_weight(event)
    local current_track = get_current_track()
    if not current_track then return end

    local mus = (BeardLib.MusicMods and BeardLib.MusicMods[current_track]) or nil
    local mus_event = (mus and mus.events and mus.events[event]) or nil

    if mus_event then
        for _, mus_track in pairs(mus_event.tracks) do
            local target_found = self:IsTargetSource(mus_track.source.path)
            if target_found then
                mus_track.weight = mus_track.orig_weight - easeInExpo(0, mus_track.orig_weight, math.min(get_time()/self.settings.time_to_sweet, 1))
                if mus_track.weight <= 0 then
                    -- maybe we can switch out of this
                    mus_track.allow_switch = true
                end
            end
        end
    end
end

function D_RUNNIN:janitorialize_weights(event)
    local current_track = get_current_track()
    if not current_track then return end

    local mus = (BeardLib.MusicMods and BeardLib.MusicMods[current_track]) or nil
    local mus_event = (mus and mus.events and mus.events[event]) or nil

    --[[
        BeardLib thinks all its tracks come via load time
        and so it doesn't believe in any strangely weighted items
        ending up in its lap

        see: [additional reading](https://youtu.be/E80RcR9VOyY?t=120)
    ]]
    local dead_tracks = {}
    if mus_event then
        for _, mus_track in pairs(mus_event.tracks) do
            if mus_track.weight <= 0 then
                -- log('D_RUNNIN: killing' .. mus_track.source.path  .. ' at ' .. tostring(mus_track.weight) .. ' for ' .. tostring(get_time()))
                dead_tracks[mus_track] = true
            end
        end

        for i = #mus_event.tracks, 1, -1 do
            if dead_tracks[mus_event.tracks[i]] then
                table.remove(mus_event.tracks, i)
            end
        end
    end
end

-- local ops = Hooks:GetFunction(MusicManager, 'post_event')
---@diagnostic disable-next-line: unused-local
local function repost_event(self, name)
    D_RUNNIN.think_dump = true
    -- tracer(name, ...)
    local event = bare_event(name or '[sorry nothing]')
    if event then
        local current_track = get_current_track()
        if current_track then
            -- first time running latch
            -- local first_run = not D_RUNNIN.did_first_run
            if not D_RUNNIN.did_first_run then
                D_RUNNIN:setup_tracks()
                D_RUNNIN:JarCandy()

                D_RUNNIN.did_first_run = true
            end

            D_RUNNIN:DumpCandyJar(current_track)
            D_RUNNIN:mutate_target_weight(event)

            D_RUNNIN:mutate_interlude_start(event)
            D_RUNNIN:mutate_interlude(event)
            D_RUNNIN:mutate_switch_toggle(event)
            D_RUNNIN:mutate_ambient_weight(event)

            D_RUNNIN:janitorialize_weights(event)

            -- if first_run or true then
                -- if first_run then log('DR_UNNIN: first run metrics') else log('DR_UNNIN: any time metrics for fun!!!') end
                -- local mus = BeardLib.MusicMods and BeardLib.MusicMods[ current_track ] or nil
                -- Utils.PrintTable(mus, 6)
            -- end
        -- else
            -- log('D_RUNNIN_X: ran without current track? ' .. tostring(event) ..':'.. tostring(current_track))
        end
    -- else
        -- log('D_RUNNIN_X: what is a '..tostring(event))
    end
end

Hooks:PreHook(MusicManager, 'post_event', 'D_RUNNIN_post_event_MusicManager', repost_event)

function D_RUNNIN:think_ping(t)
    local flat_time = math.floor(t)

    local not_first_time = flat_time >= self.settings.ping
    local is_time = (flat_time % self.settings.ping) == 0
    local not_same_time = self.last_ping ~= flat_time

    if is_time and not_first_time and not_same_time then
        -- log("pingle time: " .. tostring(flat_time) .. ":" .. format_round_digits(t, 2) .. '/' .. format_round_digits(self.settings.ping, 2))
        self.last_ping = flat_time
        self:ping()
    end
end

BeardLib:AddUpdater("d_runnin_inc",function(t)
    D_RUNNIN:think_ping(t)

    -- local current_track = get_current_track()
    -- local mus = BeardLib.MusicMods and BeardLib.MusicMods[current_track] or nil

    -- if D_RUNNIN.think_dump then
        -- chat_log('D_RUNNIN_THINK_DUMP:', t, dt, nt, current_track)
        -- D_RUNNIN.think_dump = false
    -- end
end)